#!/bin/bash

echo $'========================\nDumpPCIRom.sh log:\n' 1>&2

script_name=$0
script_full_path=$(dirname "$0")

echo "# Working Path: "$(pwd) 1>&2
echo "# Script Path: "$script_full_path 1>&2

detokoptions=""
if [[ $1 == "-v" ]]; then
	detokoptions="-v"
	shift
fi

if [ $# -ne 1 ] ; then
	echo "### Error in parameter list" 1>&2
	echo "### Usage - DumpPCIRom [-v] RomDataForkFile > output ‚â• messages" 1>&2
	echo "#" 1>&2
	echo "#         -v   output fcode numbers for all tokens" 1>&2
	exit 1
fi

# 1='Work:Programming:CW Projects:DumpNameRegistry:ROM Radeon 7000:ROMs:Radeon 7000 v208'
# 1="${theDir}Radeon7000v208.fc"
theROM="$1"
RomStart=0
RomFileSize=$(wc -c < "${theROM}")

detokErrors="/tmp/detokErrors"
echo -n > "${detokErrors}"

while true; do
	#========================================
	# Dump PCI header
	
	vpd=$(dd if="${theROM}" bs=1 skip=${RomStart} count=28 2> /dev/null | xxd -g 4 -c 28 | tr '[:lower:]' '[:upper:]')
	#echo "# got vpd: '$vpd'" 1>&2
	
	pat='^........: (....)(..)(..) (.{44}) (..)(..)(....)  ....(.{20})....'
	if [[ ${vpd} =~ $pat ]] ; then
		PciMagicNumber=${BASH_REMATCH[1]}
		offset2=0
		codeTypeString="Open Firmware"
		IndicatorString="last image"
	else
		echo "# bad vpd: '$vpd'" 1>&2
		exit 1
	fi
	
	if [[ ${PciMagicNumber} == 55AA ]] ; then
		ProcessorArchitecture=${BASH_REMATCH[3]}${BASH_REMATCH[2]}
		ProcessorArchitectureUniqueData="${BASH_REMATCH[4]} ${BASH_REMATCH[8]}"
		PointerPCIDataStructure=${BASH_REMATCH[6]}${BASH_REMATCH[5]}
		PadBytes=${BASH_REMATCH[7]}
	
		pds=$(dd if="${theROM}" bs=1 skip=$((${RomStart} + 0x${PointerPCIDataStructure})) count=24 2> /dev/null | xxd -g 4 -c 24 | tr '[:lower:]' '[:upper:]')
		#echo "# got pds: '$pds'" 1>&2

		pat='^........: .{8} (..)(..)(..)(..) (..)(..)(..)(..) (..)(.{30}).{20}'
		if [[ ${pds} =~ $pat ]] ; then
			VendorID=${BASH_REMATCH[2]}${BASH_REMATCH[1]}
			DeviceID=${BASH_REMATCH[4]}${BASH_REMATCH[3]}
			PointerVitalProductData=${BASH_REMATCH[6]}${BASH_REMATCH[5]}
			PCIDataStructureLength=${BASH_REMATCH[8]}${BASH_REMATCH[7]}
			PCIDataStructureRevision=${BASH_REMATCH[9]}
			PCIInfo=${BASH_REMATCH[10]}
			#echo "# got PCIInfo: '$PCIInfo'" 1>&2
		else
			echo "# bad pds: '$pds'" 1>&2
			exit 1
		fi
	
		pat='(..)(..)(..) (..)(..)(..)(..) (..)(..)(....  ....)'
		if [[ $PCIInfo =~ $pat ]] ; then
			ClassCode=${BASH_REMATCH[3]}${BASH_REMATCH[2]}${BASH_REMATCH[1]}
			ImageLength=${BASH_REMATCH[5]}${BASH_REMATCH[4]}
			: $(( ImageBytes = 0x${ImageLength} * 512 ))
			ImageBytesHex=$(printf "%X" $ImageBytes)
			RevisionLevel=${BASH_REMATCH[7]}${BASH_REMATCH[6]}
			CodeType=${BASH_REMATCH[8]}
			Indicator=${BASH_REMATCH[9]}
			SignatureStuff=${BASH_REMATCH[10]}
			#echo "# got SignatureStuff: '$SignatureStuff'" 1>&2
		else
			echo "# bad ClassCode" 1>&2
			exit 1
		fi
	
		pat='(....)  (....)'
		if [[ $SignatureStuff =~ $pat ]] ; then
			Reserved=${BASH_REMATCH[1]}
			SignatureString=${BASH_REMATCH[2]}
		else
			echo "# bad SignatureString" 1>&2
			exit 1
		fi
	
		: $(( offset2 = ${RomStart} + 0x${PointerPCIDataStructure} + 0x${PCIDataStructureLength}))
	
		if [[ "${CodeType}" == "00" ]] ; then
			codeTypeString="BIOS"
		elif [[ "${CodeType}" == "01" ]] ; then
			codeTypeString="Open Firmware"
		elif [[ "${CodeType}" == "02" ]] ; then
			codeTypeString="Hewlett-Packard PA RISC"
		elif [[ "${CodeType}" == "03" ]] ; then
			codeTypeString="EFI"
		else
			codeTypeString="??"
		fi

		if [[ "${Indicator}" == "80" ]] ; then
			IndicatorString="last image"
		elif [[ "${Indicator}" == "00" ]] ; then
			IndicatorString="another image"
		else
			IndicatorString="??"
		fi
	
		if [[ ${PointerPCIDataStructure} != 001C ]] ; then
			dump1=$((${RomStart} + 28))
			varBytes="\                            Variable length pad bytes: 
"$(dd if="${theROM}" bs=1 skip=$dump1 count=$((0x${PointerPCIDataStructure} - 28)) 2> /dev/null | xxd -o $dump1 -c 32 -g 4 | tr '[:lower:]' '[:upper:]' | sed -E '/(.*)/s//\\\        \1/')
		else
			varBytes=""
		fi
		#echo "# got varBytes: '$varBytes'" 1>&2
	
		if [[ ${PCIDataStructureLength} != 0018 ]] ; then
			dump1=$((${RomStart} + 0x${PointerPCIDataStructure} + 24))
			varBytes2="\                            Variable length pad bytes: 
"$(dd if="${theROM}" bs=1 skip=$dump1 count=$((offset2 - dump1)) 2> /dev/null | xxd -o $dump1 -c 32 -g 4 | tr '[:lower:]' '[:upper:]' | sed -E '/(.*)/s//\\\        \1/')
		else
			varBytes2=""
		fi
		#echo "# got varBytes2: '$varBytes2'" 1>&2
	
		(
			echo -n \
"\   PCI expansion PROM header
\                              PCI magic number (55AA): ${PciMagicNumber}
\                   Processor architecture unique data: ${ProcessorArchitecture}
\      Reserved for processor architecture-unique data: ${ProcessorArchitectureUniqueData}
\               Pointer to start of PCI Data Structure: ${PointerPCIDataStructure}
\                                            Pad bytes: ${PadBytes}
${varBytes}
\   
\   PCI Data Structure (4 byte aligned)
\                              Signature string (PCIR): ${SignatureString}
\                         Vendor ID = config reg 00/01: ${VendorID}
\                         Device ID = config reg 02/03: ${DeviceID}
\                        Pointer to Vital Product Data: ${PointerVitalProductData}
\                            PCI Data Structure length: ${PCIDataStructureLength}
\                          PCI Data Structure revision: ${PCIDataStructureRevision}
\                     Class Code = config reg 09/0a/0b: ${ClassCode} (Class code / Subclass code / Programming interface code)
\                            Image Length in 512 bytes: ${ImageLength} (* 0x200 = 0x${ImageBytesHex})
\                          Revision Level of Code/Data: ${RevisionLevel}
\                                            Code Type: ${CodeType} = ${codeTypeString}
\                                            Indicator: ${Indicator} = ${IndicatorString}
\                                             reserved: ${Reserved}
${varBytes2}

hex

"
	
			if true ; then
				if [[ "${RevisionLevel}" != "0001" ]] ; then
					echo "tokenizer[ ${RevisionLevel} ]tokenizer set-rev-level"
				fi
	
				if [[ "${PointerVitalProductData}" != "0000" ]] ; then
					echo "tokenizer[ ${PointerVitalProductData} ]tokenizer set-vpd-offset"
				fi
	
				if [[ "${ProcessorArchitecture}" != "0034" ]] ; then
					echo "tokenizer[ ${ProcessorArchitecture} ]tokenizer pci-architecture"
				fi
	
				if [[ "${PointerPCIDataStructure}" != "001C" ]] ; then
					echo "tokenizer[ ${PointerPCIDataStructure} ]tokenizer pci-data-structure-start"
				fi
	
				if [[ "${PCIDataStructureLength}" != "0018" ]] ; then
					echo "tokenizer[ ${PCIDataStructureLength} ]tokenizer pci-data-structure-length"
				fi
	
				echo "tokenizer[ ${VendorID} ${DeviceID} ${ClassCode} ]tokenizer pci-header"
	
				echo "tokenizer[ ${ImageBytesHex} ]tokenizer rom-size"
	
				echo
			fi
		) > "/tmp/PCIHeader"
		
	fi
	
	
	#exit 0
	
	#========================================
	# Dump FCode
	
	sourceFCode="/tmp/FCode"
	if [ ${offset2} -ne 0 ] ; then
		startOffset=${offset2}
		dd if="${theROM}" of="${sourceFCode}" bs=1 skip=${startOffset} count=$((RomStart + ImageBytes - startOffset)) 2> /dev/null
		: $(( RomStart += ImageBytes ))
	else
		startOffset=${RomStart}
		dd if="${theROM}" of="${sourceFCode}" bs=1 skip=${startOffset} 2> /dev/null
		: $(( RomStart = 0x7FFFFFFF ))
	fi
	
	FCode0="/tmp/FCode0.of"
	
	if [[ "${codeTypeString}" == "Open Firmware" ]] ; then
		detok -t ${detokoptions} "${sourceFCode}" > "${FCode0}" 2>> "${detokErrors}"
		echo 1>&2
		perl $script_full_path/ConvertFCodeTokensToForth.pl "${FCode0}" ${offset2}
	else
		cat "/tmp/PCIHeader"
		xxd -o ${startOffset} -c 32 -g 4 | tr '[:lower:]' '[:upper:]' "${sourceFCode}" | sed -E '/(.*)/s//\\\        \1/'
	fi

	if (( RomStart >= RomFileSize )); then
		break
	fi

	echo $'\n\n'

done # Loop

# Remove lines that are supposed to occur in the errors output but are very numerous. This makes the other lines (which are likely supposed to be there but are not very numerous) stand out.
echo $'\n========================\ndetok log:\n' 1>&2
sed -E "/^Doing Pass 1/,/^Doing Pass 2/ { /# Defined program defined token/ d ; } ; /^Doing Pass 2/,$ { /# Redefined program defined token/ d ; }" "${detokErrors}" 1>&2

exit 0
